from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    Iterator,
    Sequence,
    SupportsIndex,
    TypeVar,
    overload,
)
from weakref import WeakValueDictionary

import numpy as np
from numpy.typing import NDArray
from psygnal import Signal, SignalGroup

from whitecanvas.layers import _legend
from whitecanvas.layers._base import DataBoundLayer, LayerEvents, PrimitiveLayer
from whitecanvas.protocols import layer_protocols as _lp
from whitecanvas.theme import get_theme
from whitecanvas.types import (
    Alignment,
    ColorType,
    Hatch,
    LineStyle,
    XYTextData,
    _Void,
)
from whitecanvas.utils.normalize import arr_color, as_any_1d_array, as_color_array
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.layers.group._collections import LayerCollectionBase

_L = TypeVar("_L", bound=PrimitiveLayer)
_void = _Void()


class FaceEvents(SignalGroup):
    """Events of the face changes."""

    color = Signal(object)
    hatch = Signal(object)


class EdgeEvents(SignalGroup):
    """Events of the edge changes."""

    color = Signal(object)
    style = Signal(object)
    width = Signal(object)


class LayerNamespace(ABC, Generic[_L]):
    _properties = ()
    events: SignalGroup
    _events_class: type[SignalGroup]

    def __init__(self, layer: _L | None = None) -> None:
        self._layer = layer
        self._instances: WeakValueDictionary[int, _L] = WeakValueDictionary()
        self.events = self.__class__._events_class()
        self.__setattr__ = self._setattr

    def __repr__(self) -> str:
        attrs = ", ".join(f"{p}={getattr(self, p)!r}" for p in self._properties)
        return f"{type(self).__name__}(layer={self._layer!r}, {attrs})"

    def __repr_pretty__(self, p, cycle):
        attrs = ",\n    ".join(f"{p}={getattr(self, p)!r}" for p in self._properties)
        p.text(f"{type(self).__name__}(\n    layer={self._layer!r}, {attrs}\n)")

    def _setattr(self, name: str, value: Any) -> None:
        raise AttributeError(f"Cannot set attribute {name!r} on {self!r}")

    @abstractmethod
    def _as_legend_info(self) -> _legend.LegendItem:
        """Get the legend information of the namespace."""


class FaceNamespace(LayerNamespace[PrimitiveLayer[_lp.HasFaces]]):
    _properties = ("color", "hatch")
    events: FaceEvents
    _events_class = FaceEvents

    @property
    @abstractmethod
    def color(self) -> NDArray[np.floating]:
        raise NotImplementedError

    @property
    @abstractmethod
    def hatch(self) -> Hatch | EnumArray[Hatch]:
        raise NotImplementedError

    @property
    @abstractmethod
    def alpha(self) -> float | NDArray[np.floating]:
        raise NotImplementedError

    @abstractmethod
    def update(self, color=None, hatch=None, alpha=1.0): ...


class EdgeNamespace(LayerNamespace[PrimitiveLayer[_lp.HasEdges]]):
    _properties = ("color", "style", "width")
    events: EdgeEvents
    _events_class = EdgeEvents

    @property
    @abstractmethod
    def color(self) -> NDArray[np.floating]:
        raise NotImplementedError

    @property
    @abstractmethod
    def width(self) -> float | NDArray[np.floating]:
        raise NotImplementedError

    @property
    @abstractmethod
    def style(self) -> LineStyle | EnumArray[LineStyle]:
        raise NotImplementedError

    @property
    @abstractmethod
    def alpha(self) -> float | NDArray[np.floating]:
        raise NotImplementedError

    @abstractmethod
    def update(self, color=None, width=None, style=None, alpha=1.0): ...


class SinglePropertyFaceBase(FaceNamespace):
    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str | _Void = _void,
        alpha: float | _Void = _void,
    ) -> _L:
        """
        Update the face properties.

        Parameters
        ----------
        color : ColorType, optional
            Color of the face.
        hatch : FacePattern, optional
            Fill hatch of the face.
        alpha : float, optional
            Alpha value of the face.
        """
        if color is not _void:
            self.color = color
        if hatch is not _void:
            self.hatch = hatch
        if alpha is not _void:
            self.alpha = alpha
        return self._layer

    def _as_legend_info(self):
        return _legend.FaceInfo(self.color, self.hatch)


class SinglePropertyEdgeBase(EdgeNamespace):
    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        style: LineStyle | str | _Void = _void,
        width: float | _Void = _void,
        alpha: float | _Void = _void,
    ) -> _L:
        if color is not _void:
            self.color = color
        if style is not _void:
            self.style = style
        if width is not _void:
            self.width = width
        if alpha is not _void:
            self.alpha = alpha
        return self._layer

    def _as_legend_info(self):
        return _legend.EdgeInfo(self.color, self.width, self.style)


class MultiPropertyFaceBase(FaceNamespace):
    def update(
        self,
        *,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        hatch: Hatch | str | Sequence[Hatch | str] | _Void = _void,
        alpha: float | Sequence[float] | _Void = _void,
    ) -> _L:
        """
        Update the face properties.

        Parameters
        ----------
        color : ColorType, optional
            Color of the face.
        hatch : FacePattern, optional
            Fill hatch of the face.
        alpha : float, optional
            Alpha value of the face.
        """
        if color is not _void:
            self.color = color
        if hatch is not _void:
            self.hatch = hatch
        if alpha is not _void:
            self.alpha = alpha
        return self._layer

    def _as_legend_info(self):
        if self.color.size == 0:
            return _legend.EmptyLegendItem()
        color = self.color[0]
        hatch = self.hatch[0]
        return _legend.FaceInfo(color, hatch)

    def _ndata(self) -> int:
        return self._layer.ndata

    @property
    def alpha(self) -> float:
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color


class MultiPropertyEdgeBase(EdgeNamespace):
    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        style: LineStyle | str | _Void = _void,
        width: float | _Void = _void,
        alpha: float | _Void = _void,
    ):
        if color is not _void:
            self.color = color
        if style is not _void:
            self.style = style
        if width is not _void:
            self.width = width
        if alpha is not _void:
            self.alpha = alpha
        return self._layer

    def _as_legend_info(self):
        if self.color.size == 0:
            return _legend.EmptyLegendItem()
        color = self.color[0]
        width = self.width[0]
        style = self.style[0]
        return _legend.EdgeInfo(color, width, style)

    def _ndata(self) -> int:
        return self._layer.ndata

    @property
    def alpha(self) -> float:
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color


class MonoFace(SinglePropertyFaceBase):
    @property
    def color(self) -> NDArray[np.floating]:
        return self._layer._backend._plt_get_face_color()

    @color.setter
    def color(self, value: ColorType):
        col = arr_color(value)
        self._layer._backend._plt_set_face_color(col)
        self.events.color.emit(col)

    @property
    def hatch(self) -> Hatch:
        return self._layer._backend._plt_get_face_hatch()

    @hatch.setter
    def hatch(self, value: str | Hatch):
        hatch = Hatch(value)
        self._layer._backend._plt_set_face_hatch(hatch)
        self.events.hatch.emit(hatch)

    @property
    def alpha(self) -> float:
        return float(self.color[3])

    @alpha.setter
    def alpha(self, value: float):
        if not 0 <= value <= 1:
            raise ValueError(f"Alpha must be between 0 and 1, got {value!r}")
        self.color = (*self.color[:3], value)


class ConstFace(SinglePropertyFaceBase):
    @property
    def color(self) -> NDArray[np.floating]:
        return self._layer._backend._plt_get_face_color()[0]

    @color.setter
    def color(self, value: ColorType):
        col = arr_color(value)
        self._layer._backend._plt_set_face_color(col)
        self.events.color.emit(col)

    @property
    def hatch(self) -> Hatch:
        return self._layer._backend._plt_get_face_hatch()[0]

    @hatch.setter
    def hatch(self, value: str | Hatch):
        hatch = Hatch(value)
        self._layer._backend._plt_set_face_hatch(hatch)
        self.events.hatch.emit(hatch)

    @property
    def alpha(self) -> float:
        return float(self.color[3])

    @alpha.setter
    def alpha(self, value: float):
        if not is_real_number(value):
            raise ValueError(
                "Alpha must be a real number for a constant-face layer. If you want to "
                "set multiple alpha values, use `with_face_multi` method to convert "
                "the layer to a multi-face layer."
            )
        if not 0 <= value <= 1:
            raise ValueError(f"Alpha must be between 0 and 1, got {value!r}")
        try:
            self.color = (*self.color[:3], value)
        except IndexError:
            pass  # when the layer is empty


class MonoEdge(SinglePropertyEdgeBase):
    @property
    def color(self) -> NDArray[np.floating]:
        return self._layer._backend._plt_get_edge_color()

    @color.setter
    def color(self, value: ColorType):
        col = arr_color(value)
        self._layer._backend._plt_set_edge_color(col)
        self.events.color.emit(col)

    @property
    def style(self) -> LineStyle:
        return self._layer._backend._plt_get_edge_style()

    @style.setter
    def style(self, value: str | LineStyle):
        style = LineStyle(value)
        self._layer._backend._plt_set_edge_style(style)
        self.events.style.emit(style)

    @property
    def width(self) -> float:
        return self._layer._backend._plt_get_edge_width()

    @width.setter
    def width(self, value: float):
        if value < 0:
            raise ValueError(f"Edge width must be non-negative, got {value!r}")
        value = float(value)
        self._layer._backend._plt_set_edge_width(value)
        self.events.width.emit(value)

    @property
    def alpha(self) -> float:
        """The alpha value of the edge."""
        return float(self.color[3])

    @alpha.setter
    def alpha(self, value: float):
        if not 0 <= value <= 1:
            raise ValueError(f"Alpha must be between 0 and 1, got {value!r}")
        self.color = (*self.color[:3], value)


class ConstEdge(SinglePropertyEdgeBase):
    _properties = ("color", "style", "width")

    @property
    def color(self) -> NDArray[np.floating]:
        colors = self._layer._backend._plt_get_edge_color()
        if len(colors) > 0:
            return colors[0]
        return np.array([0, 0, 0, 0], dtype=np.float32)

    @color.setter
    def color(self, value: ColorType):
        col = arr_color(value)
        self._layer._backend._plt_set_edge_color(col)
        self.events.color.emit(col)

    @property
    def style(self) -> LineStyle:
        """Edge style."""
        styles = self._layer._backend._plt_get_edge_style()
        if len(styles) > 0:
            return styles[0]
        return LineStyle.SOLID

    @style.setter
    def style(self, value: str | LineStyle):
        style = LineStyle(value)
        self._layer._backend._plt_set_edge_style(style)
        self.events.style.emit(style)

    @property
    def width(self) -> float:
        """Edge width."""
        widths = self._layer._backend._plt_get_edge_width()
        if len(widths) > 0:
            return widths[0]
        return 0.0

    @width.setter
    def width(self, value: float):
        if not is_real_number(value):
            raise ValueError(
                "Width must be a real number for a constant-edge layer. If you want to "
                "set multiple width values, use `with_edge_multi` method to convert "
                "the layer to a multi-edge layer."
            )
        if value < 0:
            raise ValueError(f"Edge width must be non-negative, got {value!r}")
        value = float(value)
        self._layer._backend._plt_set_edge_width(value)
        self.events.width.emit(value)

    @property
    def alpha(self) -> float:
        """The alpha value of the edge."""
        return float(self.color[3])

    @alpha.setter
    def alpha(self, value: float):
        if not is_real_number(value):
            raise ValueError(
                "Alpha must be a real number for a constant-edge layer. If you want to "
                "set multiple alpha values, use `with_edge_multi` method to convert "
                "the layer to a multi-edge layer."
            )
        if not 0 <= value <= 1:
            raise ValueError(f"Alpha must be between 0 and 1, got {value!r}")
        self.color = (*self.color[:3], value)


class MultiFace(MultiPropertyFaceBase):
    @property
    def color(self) -> NDArray[np.floating]:
        """Face color of the bar."""
        return self._layer._backend._plt_get_face_color()

    @color.setter
    def color(self, color):
        col = as_color_array(color, self._ndata())
        if self._ndata() > 0:
            self._layer._backend._plt_set_face_color(col)
        self.events.color.emit(col)

    @property
    def hatch(self) -> EnumArray[Hatch]:
        """Face fill hatches."""
        return np.asarray(self._layer._backend._plt_get_face_hatch(), dtype=object)

    @hatch.setter
    def hatch(self, hatch: str | Hatch | Iterable[str | Hatch]):
        if isinstance(hatch, str):
            hatch = Hatch(hatch)
        elif isinstance(hatch, Hatch):
            pass
        else:
            hatch = [Hatch(p) for p in hatch]
        if self._ndata() > 0:
            self._layer._backend._plt_set_face_hatch(hatch)
        self.events.hatch.emit(hatch)


class MultiEdge(MultiPropertyEdgeBase):
    @property
    def color(self) -> NDArray[np.floating]:
        """Edge color of the bar."""
        return self._layer._backend._plt_get_edge_color()

    @color.setter
    def color(self, color):
        col = as_color_array(color, self._ndata())
        if self._ndata() > 0:
            self._layer._backend._plt_set_edge_color(col)
        self.events.color.emit(col)

    @property
    def width(self) -> NDArray[np.floating]:
        """Edge widths."""
        return self._layer._backend._plt_get_edge_width()

    @width.setter
    def width(self, width: float | Iterable[float]):
        if not is_real_number(width):
            width = np.asarray(width, dtype=np.float32)
            if width.shape != (self._ndata(),):
                raise ValueError(
                    "Width must be a scalar or an array of length "
                    f"{self._ndata()}, got {width.shape!r}"
                )
        else:
            width = float(width)
        if self._ndata() > 0:
            self._layer._backend._plt_set_edge_width(width)
        self.events.width.emit(width)

    @property
    def style(self) -> EnumArray[LineStyle]:
        """Edge styles."""
        return np.asarray(self._layer._backend._plt_get_edge_style(), dtype=object)

    @style.setter
    def style(self, style: str | LineStyle | Iterable[str | LineStyle]):
        if isinstance(style, str):
            style = LineStyle(style)
        elif isinstance(style, LineStyle):
            pass
        else:
            style = [LineStyle(s) for s in style]
        if self._ndata() > 0:
            self._layer._backend._plt_set_edge_style(style)
        self.events.style.emit(style)


_NFace = TypeVar("_NFace", bound=FaceNamespace)
_NEdge = TypeVar("_NEdge", bound=EdgeNamespace)


class FaceEdgeMixinEvents(LayerEvents):
    face = Signal(object)
    edge = Signal(object)


class AbstractFaceEdgeMixin(Generic[_NFace, _NEdge]):
    events: FaceEdgeMixinEvents
    _events_class = FaceEdgeMixinEvents

    def __init__(self, face: _NFace, edge: _NEdge):
        self._face_namespace = face
        self._edge_namespace = edge

    def _init_events(self):
        self._face_namespace.events.connect(self.events.face.emit, max_args=None)
        self._edge_namespace.events.connect(self.events.edge.emit, max_args=None)
        # _face_namespace may change! _make_sure_hatch_visible should connected to
        # the layer event.
        self.events.face.connect(self._make_sure_hatch_visible, max_args=0)

    @property
    def face(self) -> _NFace:
        """The face namespace."""
        return self._face_namespace

    @property
    def edge(self) -> _NEdge:
        """The edge namespace."""
        return self._edge_namespace

    def with_face(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the face properties."""
        self.face.update(color=color, hatch=hatch, alpha=alpha)
        return self

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the edge properties."""
        if color is None:
            color = get_theme().foreground_color
        self.edge.update(color=color, style=style, width=width, alpha=alpha)
        return self

    def _make_sure_hatch_visible(self):
        pass

    color = property()

    @color.setter
    def color(self, value):
        self.face.color = value
        self.edge.color = value

    alpha = property()

    @alpha.setter
    def alpha(self, value):
        self.face.alpha = value
        self.edge.alpha = value

    if TYPE_CHECKING:

        def _as_legend_item(self) -> _legend.LegendItem: ...


class FaceEdgeMixin(AbstractFaceEdgeMixin[MonoFace, MonoEdge]):
    def __init__(self):
        super().__init__(MonoFace(self), MonoEdge(self))

    def _make_sure_hatch_visible(self):
        if self.face.hatch is not Hatch.SOLID and self.edge.width == 0:
            self.edge.width = 1
            self.edge.color = get_theme().foreground_color


class MultiFaceEdgeMixin(AbstractFaceEdgeMixin[_NFace, _NEdge]):
    """Mixin for layers with multiple faces and edges."""

    def __init__(self):
        super().__init__(ConstFace(self), ConstEdge(self))

    def with_face(
        self,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the face properties."""
        if not isinstance(self._face_namespace, ConstFace):
            self._face_namespace.events.disconnect()
            self._face_namespace = ConstFace(self)  # type: ignore
            self._face_namespace.events.connect(self.events.face.emit, max_args=None)
        self.face.update(color=color, hatch=hatch, alpha=alpha)
        return self

    def with_edge(
        self,
        color: ColorType | None = None,
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the edge properties."""
        if color is None:
            color = get_theme().foreground_color
        if not isinstance(self._edge_namespace, ConstEdge):
            self._edge_namespace.events.disconnect()
            self._edge_namespace = ConstEdge(self)  # type: ignore
            self._edge_namespace.events.connect(self.events.edge.emit, max_args=None)
        self.edge.update(color=color, style=style, width=width, alpha=alpha)
        return self

    def with_face_multi(
        self,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        hatch: str | Hatch | Sequence[str | Hatch] | _Void = _void,
        alpha: float = 1,
    ) -> Self:
        if not isinstance(self._face_namespace, MultiFace):
            self._face_namespace.events.disconnect()
            self._face_namespace = MultiFace(self)  # type: ignore
            self._face_namespace.events.connect(self.events.face.emit, max_args=None)
        self.face.update(color=color, hatch=hatch, alpha=alpha)
        return self

    def with_edge_multi(
        self,
        color: ColorType | Sequence[ColorType] | None = None,
        width: float | Sequence[float] = 1,
        style: str | LineStyle | list[str | LineStyle] = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the edge properties."""
        if color is None:
            color = get_theme().foreground_color
        if not isinstance(self._edge_namespace, MultiEdge):
            self._edge_namespace.events.disconnect()
            self._edge_namespace = MultiEdge(self)  # type: ignore
            self._edge_namespace.events.connect(self.events.edge.emit, max_args=None)
        self.edge.update(color=color, style=style, width=width, alpha=alpha)
        return self

    def _make_sure_hatch_visible(self):
        # TODO: following lines are needed, but it might be slow.
        # if isinstance(self.face, MonoFace):
        #     if self.face.hatch is Hatch.SOLID:
        #         return
        # else:
        #     if np.all(self.face.hatch == Hatch.SOLID):
        #         return
        _is_no_width = self.edge.width == 0
        if isinstance(self._edge_namespace, MultiEdge):
            if np.any(_is_no_width):
                ec = np.array(get_theme().foreground_color, dtype=np.float32)
                self.edge.width = np.where(_is_no_width, 1, self.edge.width)
                ec_old = self.edge.color
                ec_old[_is_no_width] = ec[np.newaxis]
                self.edge.color = ec_old
        else:
            if _is_no_width:
                self.edge.width = 1
                self.edge.color = get_theme().foreground_color


class CollectionFace(MultiPropertyFaceBase):
    _layer: LayerCollectionBase

    def _iter_children(self) -> Iterator[FaceEdgeMixin]:
        yield from iter(self._layer)

    @property
    def color(self) -> NDArray[np.floating]:
        """Face color of the bar."""
        cols = [layer.face.color for layer in self._iter_children()]
        return np.stack(cols, axis=0)

    @color.setter
    def color(self, color):
        layers = list(self._iter_children())
        ndata = len(layers)
        col = as_color_array(color, ndata)
        for layer, c in zip(layers, col):
            layer.face.color = c
        self.events.color.emit(col)

    @property
    def hatch(self) -> EnumArray[Hatch]:
        """Face fill hatch."""
        hatches = [layer.face.hatch for layer in self._iter_children()]
        return np.array(hatches, dtype=object)

    @hatch.setter
    def hatch(self, hatch: str | Hatch | Iterable[str | Hatch]):
        layers = list(self._iter_children())
        ndata = len(layers)
        hatches = as_any_1d_array(hatch, ndata, dtype=object)
        for layer, ptn in zip(layers, hatches):
            layer.face.hatch = ptn
        self.events.hatch.emit(hatch)

    @property
    def alpha(self) -> float:
        """Alpha value of the face."""
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color


class CollectionEdge(MultiPropertyEdgeBase):
    _layer: LayerCollectionBase

    def _iter_children(self) -> Iterator[FaceEdgeMixin]:
        yield from iter(self._layer)

    @property
    def color(self) -> NDArray[np.floating]:
        """Face colors of the collection."""
        cols = [layer.edge.color for layer in self._iter_children()]
        return np.stack(cols, axis=0)

    @color.setter
    def color(self, color):
        layers = list(self._iter_children())
        ndata = len(layers)
        col = as_color_array(color, ndata)
        for layer, c in zip(layers, col):
            layer.edge.color = c
        self.events.color.emit(col)

    @property
    def width(self) -> NDArray[np.float32]:
        """Edge widths."""
        widths = [layer.edge.width for layer in self._iter_children()]
        return np.array(widths, dtype=np.float32)

    @width.setter
    def width(self, width: float | Iterable[float]):
        layers = list(self._iter_children())
        widths = as_any_1d_array(width, len(layers), dtype=np.float32)
        for layer, w in zip(layers, widths):
            layer.edge.width = w
        self.events.width.emit(width)

    @property
    def style(self) -> EnumArray[LineStyle]:
        """Edge styles."""
        styles = [layer.edge.style for layer in self._iter_children()]
        return np.array(styles, dtype=object)

    @style.setter
    def style(self, style: str | LineStyle | Iterable[str | LineStyle]):
        layers = list(self._iter_children())
        styles = as_any_1d_array(style, len(layers), dtype=object)
        for layer, ls in zip(layers, styles):
            layer.edge.style = ls
        self.events.style.emit(style)

    @property
    def alpha(self) -> float:
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color


class CollectionFaceEdgeMixin(AbstractFaceEdgeMixin[CollectionFace, CollectionEdge]):
    def __init__(self):
        super().__init__(CollectionFace(self), CollectionEdge(self))

    def _make_sure_hatch_visible(self):
        _is_no_width = self.edge.width == 0
        if np.any(_is_no_width):
            ec = np.array(get_theme().foreground_color, dtype=np.float32)
            self.edge.width = np.where(_is_no_width, 1, self.edge.width)
            ec_old = self.edge.color
            ec_old[_is_no_width] = ec[np.newaxis]
            self.edge.color = ec_old


# just for typing
_E = TypeVar("_E", bound=Enum)


class EnumArray(Generic[_E]):
    @overload
    def __getitem__(self, key: SupportsIndex) -> _E: ...

    @overload
    def __getitem__(
        self, key: slice | NDArray[np.integer] | list[SupportsIndex]
    ) -> EnumArray[_E]: ...

    def __iter__(self) -> Iterator[_E]: ...


class FontEvents(SignalGroup):
    color = Signal(object)
    size = Signal(object)
    family = Signal(str)


class FontNamespace(LayerNamespace[PrimitiveLayer[_lp.HasText]]):
    _properties = ("color", "size", "family")
    events: FontEvents
    _events_class = FontEvents

    @property
    @abstractmethod
    def color(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def size(self):
        raise NotImplementedError

    @property
    def family(self):
        return self._layer._backend._plt_get_text_fontfamily()

    @family.setter
    def family(self, value):
        if value is None:
            value = get_theme().font.family
        if not isinstance(value, str):
            raise TypeError(f"fontfamily must be a string, got {type(value)}.")
        self._layer._backend._plt_set_text_fontfamily(value)
        self.events.family.emit(value)

    @abstractmethod
    def update(self, *, color=_void, size=_void, family=_void):
        raise NotImplementedError

    def _as_legend_info(self):
        raise NotImplementedError


class ConstFont(FontNamespace):
    @property
    def color(self):
        return self._layer._backend._plt_get_text_color()[0]

    @color.setter
    def color(self, value):
        if value is None:
            value = get_theme().foreground_color
        col = arr_color(value)
        self._layer._backend._plt_set_text_color(col)
        self.events.color.emit(col)

    @property
    def size(self):
        return self._layer._backend._plt_get_text_size()[0]

    @size.setter
    def size(self, value):
        if value is None:
            value = get_theme().font.size
        self._layer._backend._plt_set_text_size(value)
        self.events.size.emit(value)

    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        size: float | _Void = _void,
        family: str | _Void = _void,
    ):
        if color is not _void:
            self.color = color
        if size is not _void:
            self.size = size
        if family is not _void:
            self.family = family
        return self._layer


class MultiFont(FontNamespace):
    @property
    def color(self):
        return self._layer._backend._plt_get_text_color()

    @color.setter
    def color(self, value):
        if value is None:
            value = get_theme().foreground_color
        col = as_color_array(value, self._layer.ntexts)
        self._layer._backend._plt_set_text_color(col)
        self.events.color.emit(col)

    @property
    def size(self):
        return self._layer._backend._plt_get_text_size()

    @size.setter
    def size(self, value):
        if value is None:
            value = get_theme().font.size
        sizes = as_any_1d_array(value, self._layer.ntexts, dtype=np.float32)
        self._layer._backend._plt_set_text_size(sizes)
        self.events.size.emit(sizes)

    def update(
        self,
        *,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        size: float | Sequence[float] | _Void = _void,
        family: str | Sequence[str] | _Void = _void,
    ):
        if color is not _void:
            self.color = color
        if size is not _void:
            self.size = size
        if family is not _void:
            self.family = family
        return self._layer


_NFont = TypeVar("_NFont", bound=FontNamespace)


class TextMixinEvents(LayerEvents):
    face = Signal(object)
    edge = Signal(object)
    font = Signal(object)
    anchor = Signal(Alignment)
    rotation = Signal(object)


class TextMixin(
    DataBoundLayer[_lp.TextProtocol, XYTextData],
    Generic[_NFace, _NEdge, _NFont],
):
    face: _NFace
    edge: _NEdge
    _face_namespace: _NFace
    _edge_namespace: _NEdge
    _font_namespace: _NFont

    events: TextMixinEvents
    _events_class = TextMixinEvents

    def __init__(self, name: str | None = None):
        self._face_namespace = ConstFace(self)
        self._edge_namespace = ConstEdge(self)
        self._font_namespace = ConstFont(self)
        super().__init__(name=name)
        self._face_namespace.events.connect(self.events.face.emit, max_args=None)
        self._edge_namespace.events.connect(self.events.edge.emit, max_args=None)
        self._font_namespace.events.connect(self.events.font.emit, max_args=None)

    @property
    def face(self) -> _NFace:
        """Namespace of the text background face."""
        return self._face_namespace

    @property
    def edge(self) -> _NEdge:
        """Namespace of the text background edge."""
        return self._edge_namespace

    @property
    def font(self) -> _NFont:
        """Namespace of the text font."""
        return self._font_namespace

    def with_face(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the face properties."""
        if not isinstance(self._face_namespace, ConstFace):
            self._face_namespace.events.disconnect()
            self._face_namespace = ConstFace(self)  # type: ignore
            self._face_namespace.events.connect(self.events.face.emit, max_args=None)
        self.face.update(color=color, hatch=hatch, alpha=alpha)
        return self

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the edge properties."""
        if color is None:
            color = get_theme().foreground_color
        if not isinstance(self._edge_namespace, ConstEdge):
            self._edge_namespace.events.disconnect()
            self._edge_namespace = ConstEdge(self)  # type: ignore
            self._edge_namespace.events.connect(self.events.edge.emit, max_args=None)
        self.edge.update(color=color, style=style, width=width, alpha=alpha)
        return self

    def with_face_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        hatch: str | Hatch | Sequence[str | Hatch] = Hatch.SOLID,
        alpha: float = 1,
    ) -> Self:
        if not isinstance(self._face_namespace, MultiFace):
            self._face_namespace.events.disconnect()
            self._face_namespace = MultiFace(self)  # type: ignore
            self._face_namespace.events.connect(self.events.face.emit, max_args=None)
        self.face.update(color=color, hatch=hatch, alpha=alpha)
        return self

    def with_edge_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | None = None,
        width: float | Sequence[float] = 1,
        style: str | LineStyle | list[str | LineStyle] = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the edge properties."""
        if color is None:
            color = get_theme().foreground_color
        if not isinstance(self._edge_namespace, MultiEdge):
            self._edge_namespace.events.disconnect()
            self._edge_namespace = MultiEdge(self)  # type: ignore
            self._edge_namespace.events.connect(self.events.edge.emit, max_args=None)
        self.edge.update(color=color, style=style, width=width, alpha=alpha)
        return self

    def with_font(
        self,
        *,
        color: ColorType | None = None,
        size: float | None = None,
        family: str | None = None,
    ) -> Self:
        """Update the face properties."""
        if color is None:
            color = self.font.color
        if not isinstance(self._font_namespace, ConstFace):
            self._font_namespace.events.disconnect()
            self._font_namespace = ConstFace(self)  # type: ignore
            self._font_namespace.events.connect(self.events.font.emit, max_args=None)
        self.font.update(color=color, size=size, family=family)
        return self

    def with_font_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | None = None,
        size: float | Sequence[float] | None = None,
        family: str | Sequence[str] | None = None,
    ) -> Self:
        """Update the face properties."""
        if color is None:
            if isinstance(self._font_namespace, MultiFont):
                color = _void  # no need to update
            else:
                if self.ntexts == 0:
                    color = _void
                else:
                    color = self.font.color
        if not isinstance(self._font_namespace, MultiFont):
            self._font_namespace.events.disconnect()
            self._font_namespace = MultiFont(self)  # type: ignore
            self._font_namespace.events.connect(self.events.font.emit, max_args=None)
        self.font.update(color=color, size=size, family=family)
        return self



from __future__ import annotations

import numpy as np
from pyqtgraph.graphicsItems.ScatterPlotItem import Symbols
from qtpy import QtGui
from qtpy.QtCore import Qt

from whitecanvas.types import Hatch, LineStyle, Modifier, MouseButton, Symbol


def array_to_qcolor(arr: np.ndarray) -> QtGui.QColor:
    """Convert an array to a QColor."""
    if arr.dtype.kind not in "iuf":
        raise ValueError(f"Expected numeric array, got {arr.dtype}")
    rgba = [round(c) for c in (arr * 255)]
    if len(rgba) not in (3, 4):
        raise ValueError(f"Expected array of length 4, got {rgba}")
    return QtGui.QColor(*rgba)


_LINE_STYLE = {
    LineStyle.SOLID: Qt.PenStyle.SolidLine,
    LineStyle.DASH: Qt.PenStyle.DashLine,
    LineStyle.DOT: Qt.PenStyle.DotLine,
    LineStyle.DASH_DOT: Qt.PenStyle.DashDotLine,
}

_LINE_STYLE_INV = {v: k for k, v in _LINE_STYLE.items()}
_LINE_STYLE_INV[Qt.PenStyle.NoPen] = LineStyle.SOLID


def _create_symbol(coords: list[tuple[float, float]]):
    p = QtGui.QPainterPath()
    p.moveTo(*coords[0])
    for x, y in coords[1:]:
        p.lineTo(x, y)
    p.closeSubpath()
    return p


Symbols["-"] = _create_symbol([(0, -0.5), (0, 0.5)])
Symbols["|"] = _create_symbol([(-0.5, 0), (0.5, 0)])
Symbols["s"] = _create_symbol([(-0.5, -0.5), (-0.5, 0.5), (0.5, 0.5), (0.5, -0.5)])
Symbols["."] = _create_symbol([(0, -0.1), (0.1, 0), (0, 0.1), (-0.1, 0)])

_SYMBOL = {
    Symbol.STAR: "star",
    Symbol.DIAMOND: "d",
    Symbol.TRIANGLE_UP: "t1",
    Symbol.TRIANGLE_LEFT: "t3",
    Symbol.TRIANGLE_DOWN: "t",
    Symbol.TRIANGLE_RIGHT: "t2",
    Symbol.VBAR: "-",
    Symbol.HBAR: "|",
    Symbol.SQUARE: "s",
}

_SYMBOL_INV = {v: k for k, v in _SYMBOL.items()}

_PAINT_STYLE = {
    Hatch.SOLID: Qt.BrushStyle.SolidPattern,
    Hatch.HORIZONTAL: Qt.BrushStyle.HorPattern,
    Hatch.VERTICAL: Qt.BrushStyle.VerPattern,
    Hatch.CROSS: Qt.BrushStyle.CrossPattern,
    Hatch.DIAGONAL_BACK: Qt.BrushStyle.BDiagPattern,
    Hatch.DIAGONAL_FORWARD: Qt.BrushStyle.FDiagPattern,
    Hatch.DIAGONAL_CROSS: Qt.BrushStyle.DiagCrossPattern,
}

_PAINT_STYLE_INV = {v: k for k, v in _PAINT_STYLE.items()}
_PAINT_STYLE_INV[Qt.BrushStyle.NoBrush] = Hatch.SOLID

_QT_MODIFIERS_MAP = {
    Qt.KeyboardModifier.NoModifier: (),
    Qt.KeyboardModifier.ShiftModifier: (Modifier.SHIFT,),
    Qt.KeyboardModifier.ControlModifier: (Modifier.CTRL,),
    Qt.KeyboardModifier.AltModifier: (Modifier.ALT,),
    Qt.KeyboardModifier.MetaModifier: (Modifier.META,),
    Qt.KeyboardModifier.ShiftModifier
    | Qt.KeyboardModifier.ControlModifier: (Modifier.SHIFT, Modifier.CTRL),
    Qt.KeyboardModifier.ShiftModifier
    | Qt.KeyboardModifier.AltModifier: (Modifier.SHIFT, Modifier.ALT),
    Qt.KeyboardModifier.ShiftModifier
    | Qt.KeyboardModifier.MetaModifier: (Modifier.SHIFT, Modifier.META),
    Qt.KeyboardModifier.ControlModifier
    | Qt.KeyboardModifier.AltModifier: (Modifier.CTRL, Modifier.ALT),
    Qt.KeyboardModifier.ControlModifier
    | Qt.KeyboardModifier.MetaModifier: (Modifier.CTRL, Modifier.META),
    Qt.KeyboardModifier.AltModifier
    | Qt.KeyboardModifier.MetaModifier: (Modifier.ALT, Modifier.META),
}
_QT_BUTTON_MAP = {
    Qt.MouseButton.NoButton: MouseButton.NONE,
    Qt.MouseButton.LeftButton: MouseButton.LEFT,
    Qt.MouseButton.RightButton: MouseButton.RIGHT,
    Qt.MouseButton.MiddleButton: MouseButton.MIDDLE,
    Qt.MouseButton.BackButton: MouseButton.BACK,
    Qt.MouseButton.ForwardButton: MouseButton.FORWARD,
}


def from_qt_modifiers(qt_modifiers: Qt.KeyboardModifier) -> tuple[Modifier, ...]:
    if (modifiers := _QT_MODIFIERS_MAP.get(qt_modifiers, None)) is None:
        # NOTE: some OS have default modifiers
        _lst = []
        if Qt.KeyboardModifier.ShiftModifier & qt_modifiers:
            _lst.append(Modifier.SHIFT)
        if Qt.KeyboardModifier.ControlModifier & qt_modifiers:
            _lst.append(Modifier.CTRL)
        if Qt.KeyboardModifier.AltModifier & qt_modifiers:
            _lst.append(Modifier.ALT)
        if Qt.KeyboardModifier.MetaModifier & qt_modifiers:
            _lst.append(Modifier.META)
        modifiers = tuple(_lst)
        _QT_MODIFIERS_MAP[qt_modifiers] = modifiers
    return modifiers


def from_qt_button(qt_button: Qt.MouseButton) -> MouseButton:
    return _QT_BUTTON_MAP.get(qt_button, MouseButton.NONE)


def from_qt_line_style(style: Qt.PenStyle) -> LineStyle:
    return _LINE_STYLE_INV[style]


def to_qt_line_style(style: LineStyle) -> Qt.PenStyle:
    return _LINE_STYLE[style]


def from_qt_symbol(symbol: str) -> Symbol:
    out = _SYMBOL_INV.get(symbol, None)
    if out is None:
        return Symbol(symbol)
    return out


def to_qt_symbol(symbol: Symbol) -> str:
    return _SYMBOL.get(symbol, symbol.value)


def from_qt_brush_style(hatch: Qt.BrushStyle) -> Hatch:
    return _PAINT_STYLE_INV[hatch]


def to_qt_brush_style(hatch: Hatch) -> Qt.BrushStyle:
    # BUG: pyqtgraph does not support setting brush style correctly
    return _PAINT_STYLE[hatch]



from whitecanvas import new_canvas


def main():
    canvas = new_canvas(backend="matplotlib")
    x = [1, 2, 3, 4, 5]
    y0 = [2, 3, 2, 1, 2]
    y1 = [3, 2, 1, 2, 3]
    canvas.add_line(x, y0, width=3, color="red")
    canvas.add_line(x, y1, width=3, color="blue", style="--")
    canvas.show()

if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Iterator

from cmap import Color, Colormap


class ColorPalette:
    def __init__(
        self,
        cmap: str | Colormap | ColorPalette,
    ):
        if isinstance(cmap, ColorPalette):
            _cmap = cmap._cmap
            pos = cmap._pos
        else:
            _cmap = Colormap(cmap)
            if isinstance(cmap, list):
                num = len(cmap)
                pos = [i / (num - 1) for i in range(num)] + [1.0]
            elif _cmap.category in ("qualitative", "categorical"):
                num = len(_cmap.color_stops)
                pos = [(i + 0.5) / num for i in range(num)]
            else:
                pos = [0, 0.3, 0.6, 0.9, 0.2, 0.5, 0.8, 0.1, 0.4, 0.7, 1.0]
        self._cmap: Colormap = _cmap
        self._pos: list[float] = pos
        self._n_generated = 0

    @property
    def ncolors(self) -> int:
        return len(self._pos)

    def next(self, update: bool = True) -> Color:
        """Generate the next color."""
        pos = self._pos[self._n_generated % self.ncolors]
        if update:
            self._n_generated += 1
        return self._cmap(pos)

    def init(self):
        """Initialize the palette."""
        self._n_generated = 0

    def nextn(self, num: int, update: bool = True) -> list[Color]:
        """Generate the next n colors."""
        current = self._n_generated
        out = [self.next(update=True) for _ in range(num)]
        if not update:
            self._n_generated = current
        return out

    def copy(self) -> ColorPalette:
        """Make a copy of the palette."""
        return ColorPalette(self._cmap)

    def __iter__(self) -> Iterator[Color]:
        yield from iter(self.nextn(self.ncolors, update=False))



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from matplotlib.quiver import Quiver
from numpy.typing import NDArray

from whitecanvas.backend.matplotlib._base import FakeAxes, MplLayer
from whitecanvas.protocols import VectorsProtocol, check_protocol
from whitecanvas.types import LineStyle

if TYPE_CHECKING:
    from whitecanvas.backend.matplotlib.canvas import Canvas


@check_protocol(VectorsProtocol)
class Vectors(Quiver, MplLayer):
    def __init__(self, x0, dx, y0, dy):
        super().__init__(
            FakeAxes(), x0, y0, dx, dy, angles="xy", scale_units="xy", scale=1
        )

    def _plt_get_data(self):
        return self.X, self.U, self.Y, self.V

    def _plt_set_data(self, x0, dx, y0, dy):
        self.XY = np.column_stack([x0, y0])
        self.X, self.Y = x0, y0
        self.set_offsets(self.XY)
        self.set_UVC(dx, dy)

    def post_add(self, canvas: Canvas):
        self.transform = canvas._axes.transData
        self.set_offset_transform(canvas._axes.transData)
        self._axes = canvas._axes

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> float:
        return self.get_linewidth()

    def _plt_set_edge_width(self, width: float):
        self.set_linewidth(width)

    def _plt_get_edge_style(self) -> LineStyle:
        return LineStyle(self.get_linestyle())

    def _plt_set_edge_style(self, style: LineStyle):
        self.set_linestyle(style.value)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.get_edgecolor()

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.set_color(color)

    def _plt_get_antialias(self) -> bool:
        return self.get_antialiased()

    def _plt_set_antialias(self, antialias: bool):
        self.set_antialiased(antialias)



from __future__ import annotations

from typing import Any, Generic, Iterable, TypeVar

import numpy as np
from numpy.typing import NDArray

from whitecanvas.layers._base import Layer, LayerWrapper
from whitecanvas.layers.tabular import _jitter
from whitecanvas.layers.tabular._df_compat import DataFrameWrapper, parse

_L = TypeVar("_L", bound="Layer")
_DF = TypeVar("_DF")


class DataFrameLayerWrapper(LayerWrapper[_L], Generic[_L, _DF]):
    def __init__(self, base: _L, source: DataFrameWrapper[_DF]):
        super().__init__(base)
        self._source = source

    @property
    def data(self) -> _DF:
        """The internal dataframe."""
        return self._source.get_native()


class ColumnOrValue:
    def __init__(self, by, df: DataFrameWrapper[_DF]):
        if isinstance(by, str):
            if by in df.iter_keys():
                self._is_columns = True
                self._value = (by,)
            else:
                self._is_columns = False
                self._value = by
        elif hasattr(by, "__iter__"):
            self._is_columns = all(isinstance(each, str) for each in by)
            if self._is_columns:
                columns = set(df.iter_keys())
                for each in by:
                    if each not in columns:
                        raise ValueError(f"{each!r} is not a column.")
                self._value = tuple(by)
            else:
                self._value = by
        else:
            self._is_columns = False
            self._value = by

    @property
    def is_column(self) -> bool:
        """True if the value is column name(s)."""
        return self._is_columns

    @property
    def value(self) -> Any:
        """Return the value."""
        return self._value

    @property
    def columns(self) -> tuple[str, ...]:
        """Return the column name(s)."""
        if self._is_columns:
            return self._value
        else:
            raise ValueError("The value is not a column name(s).")


def join_columns(
    *objs: Any | Iterable[Any] | None,
    source: DataFrameWrapper[_DF],
) -> tuple[str, ...]:
    """Join objects if it is a column name of the source data frame."""
    out = []
    for obj in objs:
        if obj is None:
            continue
        cv = ColumnOrValue(obj, source)
        if cv.is_column:
            for each in cv.columns:
                if each not in out:
                    out.append(each)
    return tuple(out)


def unique_tuple(a: tuple[str, ...], b: tuple[str, ...]) -> tuple[str, ...]:
    b_filt = tuple(x for x in b if x not in a)
    return a + b_filt


def norm_dodge(
    source: DataFrameWrapper[_DF],
    offset: str | tuple[str, ...],
    *args: str | tuple[str, ...] | None,
    dodge: str | tuple[str, ...] | bool = False,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    if isinstance(offset, str):
        offset = (offset,)
    if isinstance(dodge, bool):
        if dodge:
            _all = join_columns(*args, source=source)
            dodge = tuple(c for c in _all if c not in offset)
        else:
            dodge = ()
    elif isinstance(dodge, str):
        dodge = (dodge,)
    else:
        dodge = tuple(dodge)
    splitby = join_columns(offset, *args, dodge, source=source)
    return splitby, dodge


def norm_dodge_markers(
    source: DataFrameWrapper[_DF],
    offset: str | tuple[str, ...],
    color: str | tuple[str, ...] | None = None,
    hatch: str | tuple[str, ...] | None = None,
    dodge: str | tuple[str, ...] | bool = False,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    if isinstance(offset, str):
        offset = (offset,)
    if isinstance(dodge, bool):
        if dodge:
            _all = join_columns(color, hatch, source=source)
            dodge = tuple(c for c in _all if c not in offset)
        else:
            dodge = ()
    elif isinstance(dodge, str):
        dodge = (dodge,)
    else:
        dodge = tuple(dodge)
    splitby = join_columns(offset, dodge, source=source)
    return splitby, dodge


def list_to_df(arr: list[tuple], columns: list[str]) -> DataFrameWrapper[dict]:
    """
    Convert a list of tuples to a DataFrame.

    >>> list_to_df([(1, 2), (3, 4)], ["a", "b"])
    # Out: DataFrameWrapper({"a": [1, 3], "b": [2, 4]})
    """
    out = {c: [] for c in columns}
    for row in arr:
        for c, v in zip(columns, row):
            out[c].append(v)
    return parse(out)


def resolve_stacking(
    df: DataFrameWrapper[_DF],
    x: str | _jitter.JitterBase,
    y: str | _jitter.JitterBase,
    stackby: tuple[str, ...],
) -> tuple[NDArray[np.number], NDArray[np.number], NDArray[np.number]]:
    if isinstance(x, _jitter.JitterBase):
        xj = x
    else:
        xj = _jitter.IdentityJitter(x)
    if isinstance(y, _jitter.JitterBase):
        yj = y
    else:
        yj = _jitter.IdentityJitter(y)
    # pre-calculate all the possible xs
    all_x = xj.map(df)

    def _hash_rule(x: float) -> int:
        return int(round(x * 1000))

    ycumsum = {_hash_rule(_x): 0.0 for _x in all_x}
    x0 = list[NDArray[np.number]]()
    y0 = list[NDArray[np.number]]()
    b0 = list[NDArray[np.number]]()
    for _, sub in df.group_by(stackby):
        this_x = xj.map(sub)
        this_h = yj.map(sub)
        bottom = []
        for _x, _h in zip(this_x, this_h):
            _x_hash = _hash_rule(_x)
            dy = ycumsum[_x_hash]
            bottom.append(dy)
            ycumsum[_x_hash] += _h
        x0.append(this_x)
        y0.append(this_h)
        b0.append(bottom)
    x0 = np.concatenate(x0)
    y0 = np.concatenate(y0)
    b0 = np.concatenate(b0)
    return x0, y0, b0



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator

import numpy as np
from numpy.typing import NDArray
from psygnal import Signal, SignalGroup
from typing_extensions import override

from whitecanvas import protocols
from whitecanvas.backend import Backend
from whitecanvas.canvas import Canvas, CanvasBase
from whitecanvas.canvas._linker import link_axes
from whitecanvas.theme import get_theme
from whitecanvas.utils.normalize import arr_color

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.types import ColormapType


class GridEvents(SignalGroup):
    drawn = Signal()


class CanvasGrid:
    _CURRENT_INSTANCE: CanvasGrid | None = None
    events: GridEvents

    def __init__(
        self,
        heights: list[int],
        widths: list[int],
        *,
        backend: Backend | str | None = None,
    ) -> None:
        self._heights = heights
        self._widths = widths
        self._backend = Backend(backend)
        self._backend_object = self._create_backend()
        self._canvas_array = np.empty((len(heights), len(widths)), dtype=object)
        self._canvas_array.fill(None)

        # link axes
        self._x_linked = False
        self._y_linked = False
        self._x_linker_ref = None
        self._y_linker_ref = None

        # update settings
        theme = get_theme()
        self.background_color = theme.background_color
        self.size = theme.canvas_size
        self.events = GridEvents()
        self.__class__._CURRENT_INSTANCE = self

    @property
    def shape(self) -> tuple[int, int]:
        """The (row, col) shape of the grid"""
        return self._canvas_array.shape

    def link_x(self, *, future: bool = True, hide_ticks: bool = True) -> Self:
        """
        Link all the x-axes of the canvases in the grid.

        >>> from whitecanvas import new_grid
        >>> g = new_grid(2, 2).link_x()  # link x-axes of all canvases

        Parameters
        ----------
        future : bool, default True
            If Ture, all the canvases added in the future will also be linked. Only link
            the existing canvases if False.
        """
        if self._x_linker_ref is not None:
            self._x_linker_ref.unlink_all()  # initialize linker
        to_link = []
        for (_r, _), _canvas in self._iter_canvas():
            to_link.append(_canvas.x)
            if hide_ticks and _r != self.shape[0] - 1:
                _canvas.x.ticks.visible = False
        self._x_linker_ref = link_axes(to_link)
        if future:
            self._x_linked = True
            if hide_ticks:
                self._backend_object._plt_set_spacings(6, 6)
        return self

    def link_y(self, *, future: bool = True, hide_ticks: bool = True) -> Self:
        """
        Link all the y-axes of the canvases in the grid.

        >>> from whitecanvas import new_grid
        >>> g = new_grid(2, 2).link_y()  # link y-axes of all canvases

        Parameters
        ----------
        future : bool, default True
            If Ture, all the canvases added in the future will also be linked. Only link
            the existing canvases if False.
        """
        if self._y_linker_ref is not None:
            self._y_linker_ref.unlink_all()
        to_link = []
        for (_, _c), _canvas in self._iter_canvas():
            to_link.append(_canvas.y)
            if hide_ticks and _c != 0:
                _canvas.y.ticks.visible = False
        self._y_linker_ref = link_axes(to_link)
        if future:
            self._y_linked = True
            if hide_ticks:
                self._backend_object._plt_set_spacings(6, 6)
        return self

    def __repr__(self) -> str:
        cname = type(self).__name__
        w, h = self._size
        hex_id = hex(id(self))
        return f"<{cname} ({w:.1f} x {h:.1f}) at {hex_id}>"

    def __getitem__(self, key: tuple[int, int]) -> Canvas:
        canvas = self._canvas_array[key]
        if canvas is None:
            raise ValueError(f"Canvas at {key} is not set")
        elif isinstance(canvas, np.ndarray):
            raise ValueError(f"Cannot index by {key}.")
        return canvas

    def _create_backend(self) -> protocols.CanvasGridProtocol:
        return self._backend.get("CanvasGrid")(
            self._heights, self._widths, self._backend._app
        )

    def fill(self, palette: ColormapType | None = None) -> Self:
        """Fill the grid with canvases."""
        for _ in self._iter_add_canvas(palette=palette):
            pass
        return self

    def add_canvas(
        self,
        row: int,
        col: int,
        rowspan: int = 1,
        colspan: int = 1,
        *,
        palette: str | None = None,
    ) -> Canvas:
        """Add a canvas to the grid at the given position"""
        for idx, item in np.ndenumerate(self._canvas_array[row, col]):
            if item is not None:
                raise ValueError(f"Canvas already exists at {idx}")
        backend_canvas = self._backend_object._plt_add_canvas(
            row, col, rowspan, colspan
        )
        canvas = self._canvas_array[row, col] = Canvas.from_backend(
            backend_canvas,
            backend=self._backend,
            palette=palette,
        )
        # Now backend axes/viewbox are created, we can install mouse events
        canvas._install_mouse_events()

        # link axes if needed
        if self._x_linked:
            self._x_linker_ref.link(canvas.x)
        if self._y_linked:
            self._y_linker_ref.link(canvas.y)
        canvas.events.drawn.connect(self.events.drawn.emit, unique=True, max_args=None)
        return canvas

    def add_canvas_3d(
        self,
        row: int,
        col: int,
        rowspan: int = 1,
        colspan: int = 1,
        *,
        palette: str | None = None,
    ):
        """Add a canvas to the grid at the given position"""
        from whitecanvas.canvas.canvas3d._base import Canvas3D

        for idx, item in np.ndenumerate(self._canvas_array[row, col]):
            if item is not None:
                raise ValueError(f"Canvas already exists at {idx}")
        backend_canvas = self._backend_object._plt_add_canvas_3d(
            row, col, rowspan, colspan
        )
        canvas = self._canvas_array[row, col] = Canvas3D.from_backend(
            backend_canvas,
            backend=self._backend,
            palette=palette,
        )
        # Now backend axes/viewbox are created, we can install mouse events
        # canvas._install_mouse_events()

        # link axes if needed
        if self._x_linked:
            self._x_linker_ref.link(canvas.x)
        if self._y_linked:
            self._y_linker_ref.link(canvas.y)
        canvas.events.drawn.connect(self.events.drawn.emit, unique=True, max_args=None)
        return canvas

    def _iter_add_canvas(self, **kwargs) -> Iterator[Canvas]:
        for row in range(len(self._heights)):
            for col in range(len(self._widths)):
                yield self.add_canvas(row, col, **kwargs)

    def _iter_canvas(self) -> Iterator[tuple[tuple[int, int], Canvas]]:
        yielded: set[int] = set()
        for idx, canvas in np.ndenumerate(self._canvas_array):
            _id = id(canvas)
            if canvas is None or _id in yielded:
                continue
            yield idx, canvas
            yielded.add(_id)

    def show(self, block=False) -> None:
        """Show the grid."""
        from whitecanvas.backend._app import get_app

        # TODO: implement other event loops
        app = get_app(self._backend._app)
        _backend_app = app.get_app()
        out = self._backend_object._plt_show()

        if out is NotImplemented:
            from whitecanvas.backend._window import view

            view(self, self._backend.app)

        if block:
            # TODO: automatically block the event loop or enable ipython
            # GUI mode if needed.
            app.run_app()

    @property
    def background_color(self) -> NDArray[np.floating]:
        """Background color of the canvas."""
        return arr_color(self._backend_object._plt_get_background_color())

    @background_color.setter
    def background_color(self, color):
        self._backend_object._plt_set_background_color(arr_color(color))

    def screenshot(self) -> NDArray[np.uint8]:
        """Return a screenshot of the grid."""
        return self._backend_object._plt_screenshot()

    @property
    def size(self) -> tuple[int, int]:
        """Size in width x height."""
        return self._size

    @size.setter
    def size(self, size: tuple[int, int]):
        w, h = size
        if w <= 0 or h <= 0:
            raise ValueError("Size must be positive")
        self._size = (int(w), int(h))
        self._backend_object._plt_set_figsize(*self._size)

    def _repr_png_(self):
        """Return PNG representation of the widget for QtConsole."""
        from io import BytesIO

        try:
            from imageio import imwrite
        except ImportError:
            return None

        rendered = self.screenshot()
        if rendered is not None:
            with BytesIO() as file_obj:
                imwrite(file_obj, rendered, format="png")
                file_obj.seek(0)
                return file_obj.read()
        return None

    def _ipython_display_(self, *args: Any, **kwargs: Any) -> Any:
        if hasattr(self._backend_object, "_ipython_display_"):
            return self._backend_object._ipython_display_(*args, **kwargs)
        raise NotImplementedError()

    def _repr_mimebundle_(self, *args: Any, **kwargs: Any) -> dict:
        if hasattr(self._backend_object, "_repr_mimebundle_"):
            return self._backend_object._repr_mimebundle_(*args, **kwargs)
        raise NotImplementedError()

    def _repr_html_(self, *args: Any, **kwargs: Any) -> str:
        if hasattr(self._backend_object, "_repr_html_"):
            return self._backend_object._repr_html_(*args, **kwargs)
        raise NotImplementedError()

    def to_html(self, file: str | None = None) -> str:
        """Return HTML representation of the grid."""
        html = self._backend.get("to_html")(self._backend_object)
        if file is not None:
            Path(file).write_text(html, encoding="utf-8")
        return html


class CanvasVGrid(CanvasGrid):
    @override
    def __init__(
        self,
        heights: list[int],
        *,
        backend: Backend | str | None = None,
    ) -> None:
        super().__init__(heights, [1], backend=backend)

    @override
    def __getitem__(self, key: int) -> Canvas:
        canvas = self._canvas_array[key, 0]
        if canvas is None:
            raise ValueError(f"Canvas at {key} is not set")
        return canvas

    @override
    def add_canvas(self, row: int, **kwargs) -> Canvas:
        return super().add_canvas(row, 0, **kwargs)

    @override
    def _iter_add_canvas(self, **kwargs) -> Iterator[Canvas]:
        for row in range(len(self._heights)):
            yield self.add_canvas(row, **kwargs)


class CanvasHGrid(CanvasGrid):
    @override
    def __init__(
        self,
        widths: list[int],
        *,
        backend: Backend | str | None = None,
    ) -> None:
        super().__init__([1], widths, backend=backend)

    @override
    def __getitem__(self, key: int) -> Canvas:
        canvas = self._canvas_array[0, key]
        if canvas is None:
            raise ValueError(f"Canvas at {key} is not set")
        return canvas

    @override
    def add_canvas(self, col: int, **kwargs) -> Canvas:
        return super().add_canvas(0, col, **kwargs)

    @override
    def _iter_add_canvas(self, **kwargs) -> Iterator[Canvas]:
        for col in range(len(self._widths)):
            yield self.add_canvas(col, **kwargs)


class _CanvasWithGrid(CanvasBase):
    def __init__(self, canvas: Canvas, grid: CanvasGrid):
        self._main_canvas = canvas
        self._grid = grid
        super().__init__(palette=canvas._color_palette)

    def _get_backend(self) -> Backend:
        """Return the backend."""
        return self._main_canvas._backend

    def _canvas(self):
        return self._main_canvas._backend_object

    @property
    def native(self) -> Any:
        """The native backend object."""
        return self._main_canvas.native

    def show(self, block: bool = False) -> None:
        """Show the canvas using the method defined in the backend."""
        self._grid.show(block=block)

    @property
    def background_color(self) -> NDArray[np.floating]:
        """Background color of the canvas."""
        return self._grid.background_color

    @background_color.setter
    def background_color(self, color):
        self._grid.background_color = color

    @property
    def size(self) -> tuple[float, float]:
        """Size of the canvas"""
        return self._grid.size

    @size.setter
    def size(self, size: tuple[float, float]):
        self._grid.size = size

    def screenshot(self) -> NDArray[np.uint8]:
        """Return a screenshot of the grid."""
        return self._grid.screenshot()

    def _repr_png_(self):
        """Return PNG representation of the widget for QtConsole."""
        return self._grid._repr_png_()

    def _repr_mimebundle_(self, *args: Any, **kwargs: Any) -> dict:
        return self._grid._repr_mimebundle_(*args, **kwargs)

    def _ipython_display_(self, *args: Any, **kwargs: Any) -> Any:
        return self._grid._ipython_display_(*args, **kwargs)

    def _repr_html_(self, *args: Any, **kwargs: Any) -> str:
        return self._grid._repr_html_(*args, **kwargs)

    def to_html(self, file: str | None = None) -> str:
        """Return HTML representation of the canvas."""
        return self._grid.to_html(file=file)


class SingleCanvas(_CanvasWithGrid):
    """
    A canvas without other subplots.

    This class is the simplest form of canvas. In `matplotlib` terms, it is a figure
    with a single axes.
    """

    def __init__(self, grid: CanvasGrid):
        if grid.shape != (1, 1):
            raise ValueError(f"Grid shape must be (1, 1), got {grid.shape}")
        self._grid = grid
        _it = grid._iter_canvas()
        _, canvas = next(_it)
        if next(_it, None) is not None:
            raise ValueError("Grid must have only one canvas")
        self._main_canvas = canvas
        super().__init__(canvas, grid)

        # NOTE: events, dims etc are not shared between the main canvas and the
        # SingleCanvas instance. To avoid confusion, the first and the only canvas
        # should be replaces with the SingleCanvas instance.
        self.mouse = grid[0, 0].mouse
        grid._canvas_array[0, 0] = self
        self.events.drawn.connect(
            self._main_canvas.events.drawn.emit, unique=True, max_args=None
        )



from __future__ import annotations

import numpy as np
import pyqtgraph as pg
from numpy.typing import NDArray
from qtpy import QtGui

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._qt_utils import (
    array_to_qcolor,
    from_qt_brush_style,
    from_qt_line_style,
    from_qt_symbol,
    to_qt_brush_style,
    to_qt_line_style,
    to_qt_symbol,
)
from whitecanvas.protocols import MarkersProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle, Symbol
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(MarkersProtocol)
class Markers(pg.ScatterPlotItem, PyQtLayer):
    def __init__(self, xdata, ydata):
        pen = QtGui.QPen(QtGui.QColor(0, 0, 0))
        pen.setCosmetic(True)
        ndata = len(xdata)
        super().__init__(
            xdata,
            ydata,
            pen=[QtGui.QPen(pen) for _ in range(ndata)],
            brush=[QtGui.QBrush(QtGui.QColor(0, 0, 0)) for _ in range(ndata)],
            antialias=False,
            useCache=False,  # NOTE: should be True eventually, but pyqtgraph has
            # a bug in caching
        )
        self.opts["tip"] = "{data}".format

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self.getData()

    def _plt_set_data(self, xdata: np.ndarray, ydata: np.ndarray):
        pens = self._get_pen()
        brushes = self._get_brush()
        ndata = xdata.size
        if ndata <= len(brushes):
            self.setData(
                xdata,
                ydata,
                pen=pens[:ndata],
                brush=brushes[:ndata],
                size=self.data["size"][:ndata],
            )
        else:
            new_pen = [QtGui.QPen(pens[-1])] * (ndata - len(pens))
            new_brush = [QtGui.QBrush(brushes[-1])] * (ndata - len(brushes))
            new_size = np.full(ndata - len(self.data["size"]), self.data["size"][-1])
            self.setData(
                xdata,
                ydata,
                pen=np.concatenate([pens, new_pen]),
                brush=np.concatenate([brushes, new_brush]),
                size=np.concatenate([self.data["size"], new_size]),
            )

    ##### HasSymbol protocol #####
    def _plt_get_symbol(self) -> Symbol:
        return from_qt_symbol(self.opts["symbol"])

    def _plt_set_symbol(self, symbol: Symbol):
        self.setSymbol(to_qt_symbol(symbol))

    def _plt_get_symbol_size(self) -> NDArray[np.floating]:
        return self.data["size"]

    def _plt_set_symbol_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(len(self.data["x"]), size)
        self.setSize(size)

    ##### HasFace protocol #####
    def _get_brush(self) -> list[QtGui.QBrush]:
        brushes = self.data["brush"]
        return brushes

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        brushes = self._get_brush()
        if len(brushes) == 0:
            return np.zeros((0, 4), dtype=np.float32)
        return np.array(
            [brush.color().getRgbF() for brush in brushes], dtype=np.float32
        )

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.data["x"]))
        brushes = self._get_brush()
        for brush, c in zip(brushes, color):
            brush.setColor(array_to_qcolor(c))
        self.setBrush(brushes)

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [from_qt_brush_style(brush.style()) for brush in self._get_brush()]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        brushes = self._get_brush()
        if isinstance(pattern, Hatch):
            ptn = to_qt_brush_style(pattern)
            for brush in brushes:
                brush.setStyle(ptn)
        else:
            for brush, ptn in zip(brushes, pattern):
                brush.setStyle(to_qt_brush_style(ptn))
        self.setBrush(brushes)

    ##### HasEdges protocol #####
    def _get_pen(self) -> list[QtGui.QPen]:
        pens = self.data["pen"]
        return pens

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        pens = self._get_pen()
        if len(pens) == 0:
            return np.zeros((0, 4), dtype=np.float32)
        return np.array([pen.color().getRgbF() for pen in pens], dtype=np.float32)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.data["x"]))
        pens = self._get_pen()
        for pen, c in zip(pens, color):
            pen.setColor(array_to_qcolor(c))
        self.setPen(pens)

    def _plt_get_edge_width(self) -> float:
        return np.array([pen.widthF() for pen in self._get_pen()], dtype=np.float32)

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_real_number(width):
            width = np.full(len(self.data["x"]), width)
        pens = self._get_pen()
        for pen, w in zip(pens, width):
            pen.setWidthF(w)
        self.setPen(pens)

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [from_qt_line_style(pen.style()) for pen in self._get_pen()]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        pens = self._get_pen()
        if isinstance(style, LineStyle):
            s = to_qt_line_style(style)
            for pen in pens:
                pen.setStyle(s)
        else:
            for pen, s in zip(pens, style):
                pen.setStyle(to_qt_line_style(s))
        self.setPen(pens)

    def _plt_connect_pick_event(self, callback):
        def cb(ins, points, ev):
            callback([p.index() for p in points])

        self.sigClicked.connect(cb)

    def _plt_set_hover_text(self, text: list[str]):
        self.data["data"] = text
        self.opts["hoverable"] = True



# In this example, two mouse events are connected to the canvas.
# 1. Dragging the left mouse button without any keyboard modifiers draws a line.
# 2. Dragging the left mouse button with the shift key held down draws a span.

import numpy as np

from whitecanvas import new_canvas
from whitecanvas.types import MouseEvent


def main():
    canvas = new_canvas("matplotlib:qt")
    x, y = np.indices((10, 10)).reshape(2, -1)
    canvas.add_markers(x, y, color="gray")

    @canvas.mouse.moved.connect
    def _add_line(e: MouseEvent):
        if e.button != "left" or e.modifiers != ():
            return
        pos0 = e.pos
        line = canvas.add_line(np.array([pos0, pos0]), color="blue")
        dragged = False
        while e.type != "release":
            data = np.array([pos0, e.pos])
            line.data = data
            yield
            dragged = True
        if not dragged:
            canvas.layers.remove(line)

    @canvas.mouse.moved.connect
    def _add_spans(e: MouseEvent):
        if e.button != "left" or e.modifiers != ("shift",):
            return
        x0 = e.pos[0]
        span = canvas.add_spans([[x0, x0]], color="pink")
        dragged = False
        while e.type != "release":
            data = np.array([[x0, e.pos[0]]])
            span.data = data
            yield
            dragged = True
        if not dragged:
            canvas.layers.remove(span)

    canvas.show(block=True)

if __name__ == "__main__":
    main()



from contextlib import contextmanager
import warnings
from cmap import Color

def assert_color_equal(a, b):
    assert Color(a) == Color(b)

def assert_color_array_equal(arr, b):
    cols = [Color(a) for a in arr]
    try:
        other = Color(b)
    except ValueError:
        other = [Color(each) for each in b]
    if isinstance(other, Color):
        ok = all([each == other for each in cols])
    else:
        if len(cols) != len(other):
            ok = False
        else:
            ok = all([a == b for a, b in zip(cols, other)])
    if not ok:
        raise AssertionError(f"Color {arr} != {b}")

@contextmanager
def filter_warning(backend: str, choices: "str | list[str]"):
    if isinstance(choices, str):
        choices = [choices]
    if backend in choices:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            yield
    else:
        yield



from __future__ import annotations

from dataclasses import Field, dataclass, field, fields
from enum import Enum

from cmap import Color, Colormap

from whitecanvas.types import Hatch, LineStyle, Symbol


@dataclass
class _BaseModel:
    mutable = True

    def __post_init__(self):
        self._set_mutable(False)

    def __setattr__(self, name, value):
        if not self.mutable:
            raise TypeError(f"{self.__class__.__name__} is immutable.")
        fld: Field = self.__class__.__dataclass_fields__[name]
        if (validator := fld.metadata.get("validator", None)) is not None:
            value = validator(value)
        super().__setattr__(name, value)

    def _set_mutable(self, value: bool = True):
        super().__setattr__("mutable", value)
        for f in fields(self):
            val = getattr(self, f.name)
            if isinstance(val, _BaseModel):
                val._set_mutable(value)

    def copy(self):
        d = {}
        for f in fields(self):
            val = getattr(self, f.name)
            if isinstance(val, _BaseModel):
                val = val.copy()
            elif isinstance(val, (bool, int, float, str, tuple, Enum)):
                pass
            elif isinstance(val, (list, dict, set)):
                val = val.copy()
            else:
                val = type(val)(val)
            d[f.name] = val
        return self.__class__(**d)

    @classmethod
    def _validate_model(cls, value):
        if isinstance(value, cls):
            return value
        elif isinstance(value, dict):
            return cls(**value)
        else:
            raise TypeError(f"Cannot convert {type(value)} to {cls.__name__}.")


def _field(obj, validator=None):
    if isinstance(obj, type):
        if issubclass(obj, _BaseModel) and validator is None:
            validator = obj._validate_model
        return field(default_factory=obj, metadata={"validator": validator})
    else:
        if isinstance(obj, (int, str, float, bool, Enum, tuple, Color)):
            if validator is None:
                validator = type(obj)
            return field(default=obj, metadata={"validator": validator})
        else:
            return field(default_factory=lambda: obj, metadata={"validator": validator})


@dataclass
class Font(_BaseModel):
    """Font of texts."""

    family: str = _field("Arial")
    size: int = _field(11)
    color: Color = _field(Color("black"))


@dataclass
class Line(_BaseModel):
    """Line style."""

    width: float = _field(2.0)
    style: LineStyle = _field(LineStyle.SOLID)


@dataclass
class Markers(_BaseModel):
    """Markers of points."""

    size: float = _field(8.0)
    hatch: Hatch = _field(Hatch.SOLID)
    symbol: Symbol = _field(Symbol.CIRCLE)


@dataclass
class Bars(_BaseModel):
    """Bar style."""

    extent: float = _field(0.8)
    hatch: Hatch = _field(Hatch.SOLID)


@dataclass
class ErrorBars(_BaseModel):
    """Error bar style."""

    width: float = _field(2.0)
    style: LineStyle = _field(LineStyle.SOLID)


def _validate_canvas_size(size) -> tuple[float, float]:
    w, h = size
    w = int(w)
    h = int(h)
    if w <= 0 or h <= 0:
        raise ValueError("Canvas size must be positive.")
    return w, h


@dataclass
class Theme(_BaseModel):
    """Plot theme."""

    font: Font = _field(Font)
    line: Line = _field(Line)
    markers: Markers = _field(Markers)
    bars: Bars = _field(Bars)
    errorbars: ErrorBars = _field(ErrorBars)
    foreground_color: Color = _field(Color("black"))
    background_color: Color = _field(Color("white"))
    canvas_size: tuple[float, float] = _field((800, 600), _validate_canvas_size)
    palette: Colormap = _field(Colormap("tab10"), Colormap)


LIGHT_THEME = Theme()
DARK_THEME = Theme(
    font=Font(color="white"),
    foreground_color="#FFFFFF",
    background_color="#000000",
    palette="tab10_light",
)
